const std = @import("std");
const mem = std.mem;
const net = std.net;
// const http = std.http;

pub fn Listener(comptime Context: type) type {
    _ = Context;
    return struct {
        // ctx: *Context,
        server: *net.Server,
        allocator: mem.Allocator,
        handler: HandlerFn,
        errorHandler: ErrorHandlerFn,

        pub const InitOptions = struct {
            address: net.Address,
            allocator: mem.Allocator,
            listen_options: net.Address.ListenOptions = .{ .reuse_address = true },
            handler: HandlerFn,
            errorHandler: ErrorHandlerFn,
        };

        pub fn init(options: InitOptions) !Self {
            var server = try options.address.listen(options.listen_options);
            return .{
                // .ctx = options.ctx,
                .server = &server,
                .allocator = options.allocator,
                .handler = options.handler,
                .errorHandler = options.errorHandler,
            };
        }

        pub fn deinit(self: Self) void {
            self.server.deinit();
        }

        pub fn listen(self: Self) !void {
            while (true) {
                const read_buf = try self.allocator.alloc(u8, 1024 * 32);
                defer self.allocator.free(read_buf);

                const connection = try self.server.accept();
                defer connection.stream.close();

                var http = std.http.Server.init(connection, read_buf);
                http: while (true) {
                    var req = http.receiveHead() catch |e| if (e == error.HttpConnectionClosing) break :http else return e;
                    const request = Request{
                        .uri = try std.Uri.parseAfterScheme("http://", req.head.target),
                        .http = &req,
                    };
                    const res: Response = self.handler(&request) catch |e| blk: {
                        break :blk self.errorHandler(&request, @errorName(e)) catch |err| {
                            try req.respond("ahoy, an internal error occurred.", .{
                                .status = .internal_server_error,
                                .reason = @errorName(err),
                            });
                            continue :http;
                        };
                    };
                    try req.respond(res.body, res.options);
                }
            }
        }

        pub const Request = struct {
            uri: std.Uri,
            http: *std.http.Server.Request,
        };

        pub const Response = struct {
            options: std.http.Server.Request.RespondOptions = .{},
            body: []const u8,
        };

        const Self = @This();

        pub const HandlerFn = *const fn (req: *const Request) anyerror!Response;
        pub const ErrorHandlerFn = *const fn (req: *const Request, err: [:0]const u8) anyerror!Response;

        // pub const Router = struct {
        //     static_routes: std.StringHashMap(HandlerFn),

        //     fallback_handler: HandlerFn,
        //     error_handler: HandlerFn,

        //     allocator: mem.Allocator,

        //     pub const Options = struct {
        //         allocator: mem.Allocator,
        //         fallback_handler: HandlerFn,
        //         error_handler: HandlerFn,
        //     };

        //     pub fn init(options: Options) Router {
        //         return .{
        //             .allocator = options.allocator,
        //             .fallback_handler = options.fallback_handler,
        //             .error_handler = options.error_handler,
        //             .static_routes = std.StringHashMap(HandlerFn).init(options.allocator),
        //         };
        //     }

        //     pub fn deinit(self: Router) void {
        //         self.static_routes.deinit();
        //     }

        //     pub fn get(self: Router, path: []const u8) HandlerFn {
        //         if (self.static_routes.get(path)) |handler_fn|
        //             return handler_fn
        //         else
        //             return self.fallback_handler;
        //     }

        //     pub fn put(self: Router, path: []const u8, handler_fn: HandlerFn) !void {
        //         if (mem.indexOfAny(u8, path, "*{}") != null) {
        //             return error.UnsupportedRoute;
        //         } else try self.static_routes.put(path, handler_fn);
        //     }

        //     pub fn remove(self: Router, path: []const u8) void {
        //         if (mem.indexOfAny(u8, path, "*{}") != null) {
        //             return;
        //         } else _ = self.static_routes.remove(path);
        //     }

        //     pub fn handler(req: Request, ctx: Router) Response {
        //         const uri = std.Uri.parseAfterScheme("http://", req.http.head.target);
        //         if (ctx.get(uri.path.percent_encoded)) |handler_fn|
        //             return handler_fn(req)
        //         else
        //             return ctx.fallback_handler(req);
        //     }
        // };
    };
}
